////////////////////////////////////////////////////////////////////////
// Source file for RGBDImage class
////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////
// Include files
////////////////////////////////////////////////////////////////////////

#include "RGBD.h"




////////////////////////////////////////////////////////////////////////
// Constructors/destructors
////////////////////////////////////////////////////////////////////////

RGBDImage::
RGBDImage(void)
  : configuration(NULL),
    configuration_index(-1),
    channels(),
    width(0), height(0),
    camera_to_world(R4Matrix(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1), 0),
    intrinsics(1,0,0, 0,1,0, 0,0,1),
    world_bbox(FLT_MAX, FLT_MAX, FLT_MAX, -FLT_MAX, -FLT_MAX, -FLT_MAX),
    opengl_texture_id(-1),
    color_filename(NULL),
    depth_filename(NULL)
{
}



RGBDImage::
RGBDImage(const char *color_filename, const char *depth_filename, const R3Matrix& intrinsics_matrix, const R4Matrix& extrinsics_matrix)
  : configuration(NULL),
    configuration_index(-1),
    channels(),
    width(0), height(0),
    camera_to_world(extrinsics_matrix, 0),
    intrinsics(intrinsics_matrix),
    world_bbox(FLT_MAX, FLT_MAX, FLT_MAX, -FLT_MAX, -FLT_MAX, -FLT_MAX),
    opengl_texture_id(-1),
    color_filename((color_filename) ? strdup(color_filename) : NULL),
    depth_filename((depth_filename) ? strdup(depth_filename) : NULL)
{
}



RGBDImage::
~RGBDImage(void)
{
  // Delete opengl texture
  if (opengl_texture_id >= 0) {
    GLuint i = opengl_texture_id;
    glDeleteTextures(1, &i);
  }
  
  // Remove from configuration
  if (configuration) {
    configuration->RemoveImage(this);
  }

  // Delete channels
  for (int i = 0; i < channels.NEntries(); i++) {
    delete channels[i];
  }

  // Delete filenames
  if (color_filename) free(color_filename);
  if (depth_filename) free(depth_filename);
}



////////////////////////////////////////////////////////////////////////
// Pixel property access functions
////////////////////////////////////////////////////////////////////////

RNRgb RGBDImage::
PixelColor(int ix, int iy) const
{
  // Return color of pixel
  if (NChannels() < 3) return RNblack_rgb;
  RNScalar r = PixelChannelValue(ix, iy, RGBD_RED_CHANNEL);
  RNScalar g = PixelChannelValue(ix, iy, RGBD_GREEN_CHANNEL);
  RNScalar b = PixelChannelValue(ix, iy, RGBD_BLUE_CHANNEL);
  return RNRgb(r, g, b);
}



R3Vector RGBDImage::
PixelWorldNormal(int ix, int iy) const
{
  // Allocate buffer of points
  const int r = 64;
  R3Point *points = new R3Point[(2*r+1) * (2*r+1)];
  int npoints = 0;
  
  // Fill buffer of points in neighborhood r of (ix, iy)
  for (int s = -r; s <= r; s++) {
    int i = ix + s;
    if ((i < 0) || (i >= width)) continue;
    for (int t = -r; t <= r; t++) {
      int j = iy + t;
      if ((j < 0) || (j >= height)) continue;
      R3Point camera_position;
      R2Point image_position(i + 0.5, j + 0.5);
      if (RGBDTransformImageToCamera(image_position, camera_position, this)) {
        points[npoints++] = camera_position;
      }
    }
  }

  // Compute normal at pixel
  if (npoints < 3) return R3zero_vector;
  R3Point centroid = R3Centroid(npoints, points);
  R3Triad triad = R3PrincipleAxes(centroid, npoints, points);
  R3Vector normal = (triad[2].Z() > 0) ? triad[2] : -triad[2];

  // Transform into world coordinates
  normal.Transform(camera_to_world);
  normal.Normalize();
  
  // Delete buffer of points
  delete [] points;

  // Return normal in world coordinates
  return normal;
}



////////////////////////////////////////////////////////////////////////
// Point propery access functions
////////////////////////////////////////////////////////////////////////

RNRgb RGBDImage::
PixelColor(const R2Point& image_position) const
{
  // Return color of pixel
  if (NChannels() < 3) return RNblack_rgb;
  RNScalar r = PixelChannelValue(image_position, RGBD_RED_CHANNEL);
  RNScalar g = PixelChannelValue(image_position, RGBD_GREEN_CHANNEL);
  RNScalar b = PixelChannelValue(image_position, RGBD_BLUE_CHANNEL);
  return RNRgb(r, g, b);
}



R3Point RGBDImage::
PixelWorldPosition(const R2Point& image_position) const
{
  // Return position of pixel in world coordinates
  R3Point world_position(0,0,0);
  RGBDTransformImageToWorld(image_position, world_position, this);
  return world_position;
}



////////////////////////////////////////////////////////////////////////
// Manipulation functions
////////////////////////////////////////////////////////////////////////

void RGBDImage::
SetPixelColor(int ix, int iy, const RNRgb& color)
{
  // Set pixel red, green, and blue
  SetPixelChannelValue(ix, iy, RGBD_RED_CHANNEL, color.R());
  SetPixelChannelValue(ix, iy, RGBD_GREEN_CHANNEL, color.G());
  SetPixelChannelValue(ix, iy, RGBD_BLUE_CHANNEL, color.B());

  // Invalidate opengl
  InvalidateOpenGL();
}



void RGBDImage::
SetPixelDepth(int ix, int iy, RNScalar depth)
{
  // Set pixel depth
  SetPixelChannelValue(ix, iy, RGBD_DEPTH_CHANNEL, depth);

  // Update bounding box
  R3Point world_position;
  R2Point image_position(ix+0.5, iy+0.5);
  if (RGBDTransformImageToWorld(image_position, world_position, this)) {
    world_bbox.Union(world_position);
  }
}



void RGBDImage::
SetPixelChannelValue(int ix, int iy, int channel_index, RNScalar value)
{
  // Set channel value at pixel
  if ((channel_index < 0) || (channel_index >= channels.NEntries())) return;
  if ((ix < 0) || (ix >= channels[channel_index]->XResolution())) return;
  if ((iy < 0) || (ix >= channels[channel_index]->YResolution())) return;
  channels[channel_index]->SetGridValue(ix, iy, value);

  // Update bounding box
  if (channel_index == RGBD_DEPTH_CHANNEL) {
    R3Point world_position;
    R2Point image_position(ix+0.5, iy+0.5);
    if (RGBDTransformImageToWorld(image_position, world_position, this)) {
      world_bbox.Union(world_position);
    }
  }

  // Invalidate opengl
  if ((channel_index >= RGBD_RED_CHANNEL) && (channel_index <= RGBD_BLUE_CHANNEL)) InvalidateOpenGL();
}



void RGBDImage::
SetChannel(int channel_index, const R2Grid& image)
{
  // Set channel
  while (channels.NEntries() <= channel_index) channels.Insert(NULL);
  R2Grid *channel = channels[channel_index];
  if (channel) delete channel;
  RNArrayEntry *entry = channels.KthEntry(channel_index);
  channels.EntryContents(entry) = new R2Grid(image);

  // Set width and height
  this->width = image.XResolution();
  this->height = image.YResolution();

  // Invalidate bounding box
  if (channel_index == RGBD_DEPTH_CHANNEL) InvalidateWorldBBox();

  // Invalidate opengl
  if ((channel_index >= RGBD_RED_CHANNEL) && (channel_index <= RGBD_BLUE_CHANNEL)) InvalidateOpenGL();
}



void RGBDImage::
SetDepthChannel(const R2Grid& image)
{
  // Set depth channel
  SetChannel(RGBD_DEPTH_CHANNEL, image);

  // Set width and height
  this->width = image.XResolution();
  this->height = image.YResolution();

  // Invalidate bounding box
  InvalidateWorldBBox();
}



void RGBDImage::
SetColorChannels(const R2Image& image)
{
  // Insert channels
  while (channels.NEntries() < RGBD_RED_CHANNEL) channels.Insert(NULL);
  if (channels.NEntries() == RGBD_RED_CHANNEL) channels.Insert(new R2Grid(image.Width(), image.Height()));
  if (channels.NEntries() == RGBD_GREEN_CHANNEL) channels.Insert(new R2Grid(image.Width(), image.Height()));
  if (channels.NEntries() == RGBD_BLUE_CHANNEL) channels.Insert(new R2Grid(image.Width(), image.Height()));

  // Copy color values
  for (int iy = 0; iy < image.Height(); iy++) {
    for (int ix = 0; ix < image.Width(); ix++) {
      RNRgb color = image.PixelRGB(ix, iy);
      channels[RGBD_RED_CHANNEL]->SetGridValue(ix, iy, color.R());      
      channels[RGBD_GREEN_CHANNEL]->SetGridValue(ix, iy, color.G());      
      channels[RGBD_BLUE_CHANNEL]->SetGridValue(ix, iy, color.B());
    }
  }

  // Set width and height
  this->width = image.Width();
  this->height = image.Height();

  // Invalidate opengl
  InvalidateOpenGL();
}



void RGBDImage::
SetCameraToWorld(const R3Affine& transformation)
{
  // Set camera_to_world transformation
  this->camera_to_world = transformation;

  // Invalidate bounding boxes
  InvalidateWorldBBox();
}



void RGBDImage::
SetExtrinsics(const R4Matrix& matrix)
{
  // Set extrinsics matrix
  this->camera_to_world.Reset(matrix);

  // Invalidate bounding box
  InvalidateWorldBBox();
}



void RGBDImage::
SetIntrinsics(const R3Matrix& matrix)
{
  // Set intrinsics matrix
  this->intrinsics = matrix;

  // Invalidate bounding box
  InvalidateWorldBBox();
}



void RGBDImage::
Transform(const R3Transformation& transformation)
{
  // Update camera_to_world transformation
  R3Affine t = R3identity_affine;
  t.Transform(transformation);
  t.Transform(camera_to_world);
  camera_to_world = t;

  // Invalidate bounding box
  InvalidateWorldBBox();
}



////////////////////////////////////////////////////////////////////////
// Draw functions
////////////////////////////////////////////////////////////////////////

void RGBDImage::
Draw(int color_scheme) const
{
  // Draw
  DrawCamera(color_scheme);
}



void RGBDImage::
DrawCamera(int color_scheme, RNLength radius) const
{
  // Get useful variables
  R3Point viewpoint = WorldViewpoint();
  R3Vector towards = WorldTowards();
  R3Vector up = WorldUp();

  // Draw color
  if (color_scheme == RGBD_PHOTO_COLOR_SCHEME) LoadColor(RGBD_INDEX_COLOR_SCHEME);
  else LoadColor(color_scheme);

  // Draw camera
  glBegin(GL_LINES);
  R3LoadPoint(viewpoint);
  R3LoadPoint(viewpoint + radius * towards);
  R3LoadPoint(viewpoint);
  R3LoadPoint(viewpoint + 0.5*radius * up);
  glEnd();
}



void RGBDImage::
DrawBBox(int color_scheme) const
{
  // Draw color
  if (color_scheme == RGBD_PHOTO_COLOR_SCHEME) LoadColor(RGBD_INDEX_COLOR_SCHEME);
  else LoadColor(color_scheme);

  // Outline bounding box
  WorldBBox().Outline();
}



void RGBDImage::
DrawImage(int color_scheme, RNLength depth) const
{
  // Get useful variables
  if ((width == 0) || (height == 0)) return;
  R3Point viewpoint = WorldViewpoint();
  R3Vector towards = WorldTowards();
  R3Vector right = WorldRight();
  R3Vector up = WorldUp();
  R3Point c = viewpoint + depth * towards;
  R3Vector dx = (depth * 0.5*width / intrinsics[0][0]) * right;
  R3Vector dy = (depth * 0.5*height / intrinsics[1][1]) * up;

  // Update/select opengl texture
  if (opengl_texture_id <= 0) ((RGBDImage *) this)->UpdateOpenGL();
  if (opengl_texture_id <= 0) return;
  glBindTexture(GL_TEXTURE_2D, opengl_texture_id);
  glEnable(GL_TEXTURE_2D);

  // Draw quad with texture coordinates
  // LoadColor(color_scheme);
  glColor3d(1.0, 1.0, 1.0);
  glBegin(GL_QUADS);
  R3LoadTextureCoords(0.0, 0.0);
  R3LoadPoint(c - dx - dy);
  R3LoadTextureCoords(1.0, 0.0);
  R3LoadPoint(c + dx - dy);
  R3LoadTextureCoords(1.0, 1.0);
  R3LoadPoint(c + dx + dy);
  R3LoadTextureCoords(0.0, 1.0);
  R3LoadPoint(c - dx + dy);
  glEnd();

  // Unselect opengl texture
  glDisable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, 0);
}



void RGBDImage::
DrawPoints(int color_scheme, int skip) const
{
  // Draw color
  LoadColor(color_scheme);

  // Draw points
  glBegin(GL_POINTS);
  R3Point world_position;
  for (int ix = 0; ix < NPixels(RN_X); ix += skip) {
    for (int iy = 0; iy < NPixels(RN_Y); iy += skip) {
      if (RGBDTransformImageToWorld(R2Point(ix+0.5, iy+0.5), world_position, this)) {
        if (color_scheme == RGBD_PHOTO_COLOR_SCHEME) RNLoadRgb(PixelColor(ix, iy));
        R3LoadPoint(world_position);
      }
    }
  }
  glEnd();
}



void RGBDImage::
LoadColor(int color_scheme) const
{
  // Check color scheme
  if (color_scheme == RGBD_INDEX_COLOR_SCHEME) {
    // Load color encoding image index
    int k = 65535.0 * (ConfigurationIndex()+1) / configuration->NImages();
    unsigned char r = 0;
    unsigned char g = (k >> 8) & 0xFF;
    unsigned char b = k & 0xFF;
    glColor3ub(r, g, b);
  }
  else if (color_scheme == RGBD_PHOTO_COLOR_SCHEME) {
    // Load white
    glColor3d(1.0, 1.0, 1.0);
  }
  else if (color_scheme == RGBD_HIGHLIGHT_COLOR_SCHEME) {
    // Load highlight color
    glColor3d(1.0, 1.0, 0.0);
  }
}



////////////////////////////////////////////////////////////////////////
// Read/write/release functions
////////////////////////////////////////////////////////////////////////

int RGBDImage::
ReadChannels(void)
{
  // Read all channels
  if (!ReadColorChannels()) return 0;
  if (!ReadDepthChannel()) return 0;
  return 1;
}



int RGBDImage::
ReadColorChannels(void)
{
  // Check filename
  if (!color_filename) return 0;

  // Get full filename
  char full_filename[4096];
  const char *dirname = (configuration) ? configuration->ColorDirectory() : NULL;
  if (dirname) sprintf(full_filename, "%s/%s", dirname, color_filename);
  else sprintf(full_filename, "%s", color_filename);

  // Read color image
  R2Image color_image;
  if (!color_image.Read(full_filename)) return 0;
  SetColorChannels(color_image);

  // Return success
  return 1;
}



int RGBDImage::
ReadDepthChannel(void)
{
  // Check filename
  if (!depth_filename) return 0;

  // Get full filename
  char full_filename[4096];
  const char *dirname = (configuration) ? configuration->DepthDirectory() : NULL;
  if (dirname) sprintf(full_filename, "%s/%s", dirname, depth_filename);
  else sprintf(full_filename, "%s", depth_filename);

  // Read depth image
  R2Grid depth_image;
  if (!depth_image.ReadFile(full_filename)) return 0;
  if (strstr(depth_filename, ".png")) depth_image.Multiply(0.001);
  SetDepthChannel(depth_image);

  // Return success
  return 1;
}



int RGBDImage::
WriteChannels(void)
{
  // Write all channels
  if (!WriteColorChannels()) return 0;
  if (!WriteDepthChannel()) return 0;
  return 1;
}



int RGBDImage::
WriteColorChannels(void)
{
  // Check filename
  if (!color_filename) return 0;
  if (NChannels() <= RGBD_BLUE_CHANNEL) return 0;

  // Get full filename
  char full_filename[4096];
  const char *dirname = (configuration) ? configuration->ColorDirectory() : NULL;
  if (dirname) sprintf(full_filename, "%s/%s", dirname, color_filename);
  else sprintf(full_filename, "%s", color_filename);

  // Compute color image
  R2Image rgb_image(width, height, 3);
  for (int iy = 0; iy < height; iy++) {
    for (int ix = 0; ix < width; ix++) {
      RNRgb rgb = PixelColor(ix, iy);
      rgb_image.SetPixelRGB(ix, iy, rgb);
    }
  }

  // Write color image
  if (!rgb_image.Write(color_filename)) return 0;

  // Return success
  return 1;
}



int RGBDImage::
WriteDepthChannel(void)
{
  // Check filename
  if (!depth_filename) return 0;
  if (NChannels() <= RGBD_DEPTH_CHANNEL) return 0;

  // Get full filename
  char full_filename[4096];
  const char *dirname = (configuration) ? configuration->DepthDirectory() : NULL;
  if (dirname) sprintf(full_filename, "%s/%s", dirname, depth_filename);
  else sprintf(full_filename, "%s", depth_filename);
  
  // Write depth image  
  R2Grid depth_image(*(channels[RGBD_DEPTH_CHANNEL]));
  if (strstr(depth_filename, ".png")) depth_image.Multiply(1000.0);
  if (!depth_image.WriteFile(depth_filename)) return 0;

  // Return success
  return 1;
}



int RGBDImage::
ReleaseChannels(void)
{
  // Release all channels
  if (!ReleaseColorChannels()) return 0;
  if (!ReleaseDepthChannel()) return 0;
  return 1;
}



int RGBDImage::
ReleaseColorChannels(void)
{
  // Write color channel before releasing it ???
  // if (!WriteColorChannels()) return 0;

  // Delete color channels
  for (int channel_index = RGBD_RED_CHANNEL; channel_index <= RGBD_BLUE_CHANNEL; channel_index++) {
    if (NChannels() <= channel_index) break;
    if (!channels[channel_index]) continue;
    delete channels[channel_index];
    RNArrayEntry *entry = channels.KthEntry(channel_index);
    channels.EntryContents(entry) = NULL;
  }

  // Return success;
  return 1;
}



int RGBDImage::
ReleaseDepthChannel(void)
{
  // Write depth channel before releasing it ???
  // if (!WriteDepthChannels()) return 0;

  // Delete depth channel
  if (NChannels() <= RGBD_DEPTH_CHANNEL) return 0;
  if (!channels[RGBD_DEPTH_CHANNEL]) return 0;
  delete channels[RGBD_DEPTH_CHANNEL];
  RNArrayEntry *entry = channels.KthEntry(RGBD_DEPTH_CHANNEL);
  channels.EntryContents(entry) = NULL;

  // Return success;
  return 1;
}



void RGBDImage::
SetColorFilename(const char *filename)
{
  // Set filename
  if (color_filename) free(color_filename);
  if (filename && strcmp(filename, "-")) color_filename = strdup(filename);
  else color_filename = NULL;
}



void RGBDImage::
SetDepthFilename(const char *filename)
{
  // Set filename
  if (depth_filename) free(depth_filename);
  if (filename && strcmp(filename, "-")) depth_filename = strdup(filename);
  else depth_filename = NULL;
}



////////////////////////////////////////////////////////////////////////
// Update functions
////////////////////////////////////////////////////////////////////////

void RGBDImage::
InvalidateWorldBBox(void)
{
  // Mark bounding box for recomputation
  world_bbox.Reset(R3Point(FLT_MAX, FLT_MAX, FLT_MAX), R3Point(-FLT_MAX, -FLT_MAX, -FLT_MAX));

  // Invalidate the configuration's bounding box
  if (configuration) configuration->InvalidateWorldBBox();
}



void RGBDImage::
UpdateWorldBBox(void)
{
  // Update bounding box
  world_bbox = R3null_box;
  for (int iy = 0; iy < height; iy++) {
    for (int ix = 0; ix < width; ix++) {
      R3Point world_position;
      R2Point image_position(ix+0.5, iy+0.5);
      if (RGBDTransformImageToWorld(image_position, world_position, this)) {
        world_bbox.Union(world_position);
      }
    }
  }
}



void RGBDImage::
InvalidateOpenGL(void) 
{
  // Delete opengl texture
  if (opengl_texture_id > 0) {
    GLuint i = opengl_texture_id;
    glDeleteTextures(1, &i);
  }

  // Reset opengl identifier
  opengl_texture_id = -1;
}


  
void RGBDImage::
UpdateOpenGL(void) 
{
  // Check identifier
  if (opengl_texture_id > 0) return;
  
  // Create identifier
  GLuint identifier;
  glGenTextures(1, &identifier);

  // Create temporary R2Image
  R2Image rgb_image(width, height, 3);
  for (int iy = 0; iy < height; iy++) {
    for (int ix = 0; ix < width; ix++) {
      rgb_image.SetPixelRGB(ix, iy, PixelColor(ix, iy));
    }
  }

  // Define texture
  glBindTexture(GL_TEXTURE_2D, identifier);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, rgb_image.Width(), rgb_image.Height(),
    GL_RGB, GL_UNSIGNED_BYTE, (const unsigned char *) rgb_image.Pixels());

  // Remember identifier
  opengl_texture_id = identifier;
}
