////////////////////////////////////////////////////////////////////////
// Source file for RGBDSurface class
////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////
// Include files
////////////////////////////////////////////////////////////////////////

#include "RGBD.h"




////////////////////////////////////////////////////////////////////////
// Constructors/destructors
////////////////////////////////////////////////////////////////////////

RGBDSurface::
RGBDSurface(void)
  : configuration(NULL),
    configuration_index(-1),
    channels(),
    width(0), height(0),
    surface_to_world(R4Matrix(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1), 0),
    resolution(0.01),
    opengl_texture_id(-1),
    texture_filename(NULL)
{
}



RGBDSurface::
RGBDSurface(const char *texture_filename,
  const R3Point& world_center, const R3Vector& world_normal, const R3Vector& world_up,
  RNLength world_radius[2], RNLength resolution)
  : configuration(NULL),
    configuration_index(-1),
    channels(),
    width(0), height(0),
    surface_to_world(R4Matrix(1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1), 0),
    resolution(resolution),
    opengl_texture_id(-1),
    texture_filename(NULL)
{
  // Set texture filename
  SetTextureFilename(texture_filename);
  
  // Compute surface-to-world transformation
  R3Vector up = world_up;
  up.Normalize();
  R3Vector normal = world_normal;
  normal.Normalize();
  R3Vector right = world_up % world_normal;
  right.Normalize();
  R3Triad triad(right, up, normal);
  R3Point origin(world_center - world_radius[0]*right - world_radius[1]*up);
  R3CoordSystem cs(origin, triad);
  surface_to_world.Reset(cs.Matrix(), 0);

  // Compute width and height from resolution
  if (resolution <= 0) resolution = 0.01;
  width = (int) (2.0 * world_radius[0] / resolution) + 1;
  height = (int) (2.0 * world_radius[1] / resolution) + 1;
  R2Affine grid_to_surface(R3Matrix(1.0/resolution, 0, 0,  0, 1.0/resolution, 0,  0, 0, 1));

  // Create channels
  R2Grid tmp(width, height, grid_to_surface);
  SetChannel(RGBD_RED_CHANNEL, tmp);
  SetChannel(RGBD_GREEN_CHANNEL, tmp);
  SetChannel(RGBD_BLUE_CHANNEL, tmp);
}



RGBDSurface::
~RGBDSurface(void)
{
  // Delete opengl texture
  InvalidateOpenGL();
  
  // Remove from configuration
  if (configuration) {
    configuration->RemoveSurface(this);
  }

  // Delete channels
  for (int i = 0; i < channels.NEntries(); i++) {
    delete channels[i];
  }

  // Delete texture filename
  if (texture_filename) free(texture_filename);
}



////////////////////////////////////////////////////////////////////////
// Pixel property access functions
////////////////////////////////////////////////////////////////////////

RNRgb RGBDSurface::
PixelColor(int ix, int iy) const
{
  // Return color of pixel
  if (NChannels() < 3) return RNblack_rgb;
  RNScalar r = PixelChannelValue(ix, iy, RGBD_RED_CHANNEL);
  RNScalar g = PixelChannelValue(ix, iy, RGBD_GREEN_CHANNEL);
  RNScalar b = PixelChannelValue(ix, iy, RGBD_BLUE_CHANNEL);
  return RNRgb(r, g, b);
}



R3Point RGBDSurface::
PixelWorldPosition(int ix, int iy) const
{
  // Return position in world coordinates
  R2Point texture_position(ix + 0.5, iy + 0.5);
  return PixelWorldPosition(texture_position);
}



R3Vector RGBDSurface::
PixelWorldNormal(int ix, int iy) const
{
  // Return surface normal
  return WorldNormal();
}



////////////////////////////////////////////////////////////////////////
// Point property access functions
////////////////////////////////////////////////////////////////////////

RNRgb RGBDSurface::
PixelColor(const R2Point& texture_position) const
{
  // Return color of pixel
  if (NChannels() < 3) return RNblack_rgb;
  RNScalar r = PixelChannelValue(texture_position, RGBD_RED_CHANNEL);
  RNScalar g = PixelChannelValue(texture_position, RGBD_GREEN_CHANNEL);
  RNScalar b = PixelChannelValue(texture_position, RGBD_BLUE_CHANNEL);
  return RNRgb(r, g, b);
}



R3Point RGBDSurface::
PixelWorldPosition(const R2Point& texture_position) const
{
  // Return position in world coordinates
  R3Point world_position(0,0,0);
  RGBDTransformTextureToWorld(texture_position, world_position, this);
  return world_position;
}



R3Vector RGBDSurface::
PixelWorldNormal(const R2Point& texture_position) const
{
  // Return surface normal
  return WorldNormal();
}



////////////////////////////////////////////////////////////////////////
// Manipulation functions
////////////////////////////////////////////////////////////////////////

void RGBDSurface::
SetPixelColor(int ix, int iy, const RNRgb& color)
{
  // Set pixel red, green, and blue
  SetPixelChannelValue(ix, iy, RGBD_RED_CHANNEL, color.R());
  SetPixelChannelValue(ix, iy, RGBD_GREEN_CHANNEL, color.G());
  SetPixelChannelValue(ix, iy, RGBD_BLUE_CHANNEL, color.B());

  // Invalidate opengl
  InvalidateOpenGL();
}



void RGBDSurface::
SetPixelChannelValue(int ix, int iy, int channel_index, RNScalar value)
{
  // Set channel value at pixel
  if ((channel_index < 0) || (channel_index >= channels.NEntries())) return;
  if ((ix < 0) || (ix >= channels[channel_index]->XResolution())) return;
  if ((iy < 0) || (ix >= channels[channel_index]->YResolution())) return;
  channels[channel_index]->SetGridValue(ix, iy, value);

  // Invalidate opengl
  if ((channel_index >= RGBD_RED_CHANNEL) && (channel_index <= RGBD_BLUE_CHANNEL)) InvalidateOpenGL();
}



void RGBDSurface::
SetChannel(int channel_index, const R2Grid& image)
{
  // Set channel
  while (channels.NEntries() <= channel_index) channels.Insert(NULL);
  R2Grid *channel = channels[channel_index];
  if (channel) delete channel;
  RNArrayEntry *entry = channels.KthEntry(channel_index);
  channels.EntryContents(entry) = new R2Grid(image);

  // Set width and height
  this->width = image.XResolution();
  this->height = image.YResolution();

  // Invalidate opengl
  if ((channel_index >= RGBD_RED_CHANNEL) && (channel_index <= RGBD_BLUE_CHANNEL)) InvalidateOpenGL();
}



void RGBDSurface::
SetColorChannels(const R2Image& image)
{
  // Insert channels
  while (channels.NEntries() < RGBD_RED_CHANNEL) channels.Insert(NULL);
  if (channels.NEntries() == RGBD_RED_CHANNEL) channels.Insert(new R2Grid(image.Width(), image.Height()));
  if (channels.NEntries() == RGBD_GREEN_CHANNEL) channels.Insert(new R2Grid(image.Width(), image.Height()));
  if (channels.NEntries() == RGBD_BLUE_CHANNEL) channels.Insert(new R2Grid(image.Width(), image.Height()));

  // Copy color values
  for (int iy = 0; iy < image.Height(); iy++) {
    for (int ix = 0; ix < image.Width(); ix++) {
      RNRgb color = image.PixelRGB(ix, iy);
      channels[RGBD_RED_CHANNEL]->SetGridValue(ix, iy, color.R());      
      channels[RGBD_GREEN_CHANNEL]->SetGridValue(ix, iy, color.G());      
      channels[RGBD_BLUE_CHANNEL]->SetGridValue(ix, iy, color.B());
    }
  }

  // Set width and height
  this->width = image.Width();
  this->height = image.Height();

  // Invalidate opengl
  InvalidateOpenGL();
  
}



void RGBDSurface::
SetSurfaceToWorld(const R3Affine& transformation)
{
  // Set matrix that transforms from surface coordinate system to world 
  this->surface_to_world = transformation;
  
  // Invalidate the configuration's bounding box
  if (configuration) configuration->InvalidateWorldBBox();
}



void RGBDSurface::
SetResolution(RNLength resolution)
{
  // Resample channels
  if (this->resolution > 0) {
    width = this->resolution * width / resolution;
    height = this->resolution * height / resolution;
    for (int i = 0; i < NChannels(); i++) {
      R2Grid *channel = Channel(i);
      channel->Resample(width, height);
    }
  }

  // Set size of a texture pixel in surface coordinates
  // This is the scale factor that transforms from texture coordinate system to surface
  this->resolution = resolution;
}



void RGBDSurface::
Transform(const R3Transformation& transformation)
{
  // Update camera_to_world transformation
  R3Affine t = R3identity_affine;
  t.Transform(transformation);
  t.Transform(surface_to_world);
  surface_to_world = t;
  
  // Invalidate the configuration's bounding box
  if (configuration) configuration->InvalidateWorldBBox();
}



////////////////////////////////////////////////////////////////////////
// Draw functions
////////////////////////////////////////////////////////////////////////

void RGBDSurface::
Draw(int color_scheme) const
{
  // Draw surface
  DrawTexture(color_scheme);
}



void RGBDSurface::
DrawPlane(int color_scheme) const
{
  // Get useful variables
  RNScalar x = Resolution() * (NPixels(RN_X)-1);
  RNScalar y = Resolution() * (NPixels(RN_Y)-1);

  // Load color
  if (color_scheme == RGBD_PHOTO_COLOR_SCHEME) LoadColor(RGBD_INDEX_COLOR_SCHEME);
  else LoadColor(color_scheme);

  // Push transformation
  SurfaceToWorld().Push();

  // Draw quad
  glBegin(GL_QUADS);
  R2LoadPoint(0.0, 0.0);
  R2LoadPoint(x, 0.0);
  R2LoadPoint(x, y);
  R2LoadPoint(0.0, y);
  glEnd();

  // Pop transformation
  SurfaceToWorld().Pop();
}



void RGBDSurface::
DrawTexture(int color_scheme) const
{
  // Get useful variables
  RNScalar x = Resolution() * (NPixels(RN_X)-1);
  RNScalar y = Resolution() * (NPixels(RN_Y)-1);

  // Update/select opengl texture
  if (opengl_texture_id <= 0) ((RGBDSurface *) this)->UpdateOpenGL();
  if (opengl_texture_id <= 0) return;
  glBindTexture(GL_TEXTURE_2D, opengl_texture_id);
  glEnable(GL_TEXTURE_2D);

  // Load color
  LoadColor(color_scheme);

  // Push transformation
  SurfaceToWorld().Push();

  // Draw quad with texture coordinates
  glBegin(GL_QUADS);
  R3LoadTextureCoords(0.0, 0.0);
  R2LoadPoint(0.0, 0.0);
  R3LoadTextureCoords(1.0, 0.0);
  R2LoadPoint(x, 0.0);
  R3LoadTextureCoords(1.0, 1.0);
  R2LoadPoint(x, y);
  R3LoadTextureCoords(0.0, 1.0);
  R2LoadPoint(0.0, y);
  glEnd();

  // Pop transformation
  SurfaceToWorld().Pop();

  // Unselect opengl texture
  glDisable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, 0);
}



void RGBDSurface::
LoadColor(int color_scheme) const
{
  // Check color scheme
  if (color_scheme == RGBD_INDEX_COLOR_SCHEME) {
    // Load color encoding index
    int k = 255.0 * (ConfigurationIndex()+1) / configuration->NSurfaces();
    unsigned char r = k;
    unsigned char g = 0;
    unsigned char b = 0;
    glColor3ub(r, g, b);
  }
  else if (color_scheme == RGBD_PHOTO_COLOR_SCHEME) {
    // Load white 
    glColor3d(1.0, 1.0, 1.0);
  }
  else if (color_scheme == RGBD_HIGHLIGHT_COLOR_SCHEME) {
    // Load highlight color
    glColor3d(1.0, 1.0, 0.0);
  }
}



////////////////////////////////////////////////////////////////////////
// Read/write/release functions
////////////////////////////////////////////////////////////////////////

int RGBDSurface::
ReadChannels(void)
{
  // Read all channels
  if (!ReadColorChannels()) return 0;
  return 1;
}



int RGBDSurface::
ReadColorChannels(void)
{
  // Check filename
  if (!texture_filename) return 0;

  // Get full filename
  char full_filename[4096];
  const char *dirname = (configuration) ? configuration->TextureDirectory() : NULL;
  if (dirname) sprintf(full_filename, "%s/%s", dirname, texture_filename);
  else sprintf(full_filename, "%s", texture_filename);
  if (!RNFileExists(full_filename)) return 0;
  
  // Read color image
  R2Image rgb_image;
  if (!rgb_image.Read(full_filename)) return 0;
  SetColorChannels(rgb_image);

  // Return success
  return 1;
}



int RGBDSurface::
WriteChannels(void)
{
  // Write all channels
  if (!WriteColorChannels()) return 0;
  return 1;
}



int RGBDSurface::
WriteColorChannels(void)
{
  // Check filename
  if (NChannels() <= RGBD_BLUE_CHANNEL) return 0;

  // Create texture filename, if there is none
  if (!texture_filename) {
    char buffer[4096];
    sprintf(buffer, "t%06d.jpg", ConfigurationIndex());
    SetTextureFilename(buffer);
  }

  // Get full filename
  char full_filename[4096];
  const char *dirname = (configuration) ? configuration->TextureDirectory() : NULL;
  if (dirname) sprintf(full_filename, "%s/%s", dirname, texture_filename);
  else sprintf(full_filename, "%s", texture_filename);

  // Compute color image
  R2Image rgb_image(width, height, 3);
  for (int iy = 0; iy < height; iy++) {
    for (int ix = 0; ix < width; ix++) {
      RNRgb rgb = PixelColor(ix, iy);
      rgb_image.SetPixelRGB(ix, iy, rgb);
    }
  }

  // Write color image
  if (!rgb_image.Write(full_filename)) return 0;

  // Return success
  return 1;
}



int RGBDSurface::
ReleaseChannels(void)
{
  // Release all channels
  if (!ReleaseColorChannels()) return 0;
  return 1;
}



int RGBDSurface::
ReleaseColorChannels(void)
{
  // Write color channel before release???
  // WriteColorChannels();
  
  // Delete color channels
  for (int channel_index = RGBD_RED_CHANNEL; channel_index <= RGBD_BLUE_CHANNEL; channel_index++) {
    if (NChannels() <= channel_index) break;
    if (!channels[channel_index]) continue;
    delete channels[channel_index];
    RNArrayEntry *entry = channels.KthEntry(channel_index);
    channels.EntryContents(entry) = NULL;
  }

  // Return success;
  return 1;
}



void RGBDSurface::
SetTextureFilename(const char *filename)
{
  // Set filename
  if (texture_filename) free(texture_filename);
  if (filename && strcmp(filename, "-")) texture_filename = strdup(filename);
  else texture_filename = NULL;
}



////////////////////////////////////////////////////////////////////////
// Update functions
////////////////////////////////////////////////////////////////////////

void RGBDSurface::
InvalidateOpenGL(void) 
{
  // Delete opengl texture
  if (opengl_texture_id > 0) {
    GLuint i = opengl_texture_id;
    glDeleteTextures(1, &i);
  }

  // Reset opengl identifier
  opengl_texture_id = -1;
}


  
void RGBDSurface::
UpdateOpenGL(void) 
{
  // Check identifier
  if (opengl_texture_id > 0) return;
  
  // Create identifier
  GLuint identifier;
  glGenTextures(1, &identifier);

  // Create temporary R2Image
  R2Image rgb_image(width, height, 3);
  for (int iy = 0; iy < height; iy++) {
    for (int ix = 0; ix < width; ix++) {
      rgb_image.SetPixelRGB(ix, iy, PixelColor(ix, iy));
    }
  }

  // Define texture
  glBindTexture(GL_TEXTURE_2D, identifier);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, rgb_image.Width(), rgb_image.Height(),
    GL_RGB, GL_UNSIGNED_BYTE, (const unsigned char *) rgb_image.Pixels());

  // Remember identifier
  opengl_texture_id = identifier;
}
