////////////////////////////////////////////////////////////////////////
// Include file for RGBDSurface class
////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////
// Class definition
////////////////////////////////////////////////////////////////////////

class RGBDSurface {
public:
  // Constructors/destructors
  RGBDSurface(void);
  RGBDSurface(const char *texture_filename, const R3Point& world_center, const R3Vector& world_normal,
    const R3Vector& world_up, RNLength world_radius[2], RNLength resolution);
  virtual ~RGBDSurface(void);

  // Configuration access functions
  RGBDConfiguration *Configuration(void) const;
  int ConfigurationIndex(void) const;
  
  // Channel access functions
  int NChannels(void) const;
  R2Grid *Channel(int channel_index) const;
  R2Grid *RedChannel(void) const;
  R2Grid *GreenChannel(void) const;
  R2Grid *BlueChannel(void) const;

  // Pixel property access functions
  int NPixels(int dim) const;
  RNRgb PixelColor(int ix, int iy) const;
  R3Point PixelWorldPosition(int ix, int iy) const;
  R3Vector PixelWorldNormal(int ix, int iy) const;
  RNScalar PixelChannelValue(int ix, int iy, int channel_index) const;

  // Point property access functions
  RNRgb PixelColor(const R2Point& texture_position) const;
  R3Point PixelWorldPosition(const R2Point& texture_position) const;
  R3Vector PixelWorldNormal(const R2Point& texture_position) const;
  RNScalar PixelChannelValue(const R2Point& texture_position, int channel_index) const;

  // Surface property functions
  R3Point WorldOrigin(void) const;
  R3Point WorldCentroid(void) const;
  R3Vector WorldNormal(void) const;
  R3Vector WorldAxis(int dimension) const;
  RNScalar WorldLength(int dimension) const;
  R3Box WorldBBox(void) const;
  RNArea WorldArea(void) const;

  // Transformation properties
  const R3Affine& SurfaceToWorld(void) const;
  RNLength Resolution(void) const;

  // Manipulation functions
  virtual void SetPixelColor(int ix, int iy, const RNRgb& color);
  virtual void SetPixelChannelValue(int ix, int iy, int channel_index, RNScalar value);
  virtual void SetChannel(int channel_index, const R2Grid& image);
  virtual void SetColorChannels(const R2Image& image);
  virtual void SetSurfaceToWorld(const R3Affine& transformation);
  virtual void SetResolution(RNLength resolution);
  virtual void Transform(const R3Transformation& transformation);

  // Draw functions
  virtual void Draw(int color_scheme = RGBD_PHOTO_COLOR_SCHEME) const;
  virtual void DrawPlane(int color_scheme = RGBD_INDEX_COLOR_SCHEME) const;
  virtual void DrawTexture(int color_scheme = RGBD_PHOTO_COLOR_SCHEME) const;
  virtual void LoadColor(int color_scheme = RGBD_PHOTO_COLOR_SCHEME) const;
  
  // Read/write functions
  virtual int ReadChannels(void);
  virtual int ReadColorChannels(void);
  virtual int WriteChannels(void);
  virtual int WriteColorChannels(void);
  virtual int ReleaseChannels(void);
  virtual int ReleaseColorChannels(void);

public:
  // Filename access functions
  const char *TextureFilename(void) const;

  // Filename manipulation functions (does not read file)
  virtual void SetTextureFilename(const char *filename);
  
  // Update functions
  virtual void InvalidateOpenGL(void);
  virtual void UpdateOpenGL(void);
  
public:
  // Internal variables
  friend class RGBDConfiguration;
  RGBDConfiguration *configuration;
  int configuration_index;
  RNArray<R2Grid *> channels;
  int width, height;
  R3Affine surface_to_world;
  RNLength resolution;
  int opengl_texture_id;
  char *texture_filename;
};



////////////////////////////////////////////////////////////////////////
// Inline functions
////////////////////////////////////////////////////////////////////////

inline RGBDConfiguration *RGBDSurface::
Configuration(void) const
{
  // Return configuration of which this surface is part
  return configuration;
}



inline int RGBDSurface::
ConfigurationIndex(void) const
{
  // Return index of this surface in configuration (i.e., where configuration->Surface(k) == this)
  return configuration_index;
}



inline int RGBDSurface::
NChannels(void) const
{
  // Return number of channels
  return channels.NEntries();
}



inline R2Grid *RGBDSurface::
Channel(int channel_index) const
{
  // Return channel
  return channels[channel_index];
}



inline R2Grid *RGBDSurface::
RedChannel(void) const
{
  // Return channel
  return channels[RGBD_RED_CHANNEL];
}



inline R2Grid *RGBDSurface::
GreenChannel(void) const
{
  // Return channel
  return channels[RGBD_GREEN_CHANNEL];
}



inline R2Grid *RGBDSurface::
BlueChannel(void) const
{
  // Return channel
  return channels[RGBD_BLUE_CHANNEL];
}



inline int RGBDSurface::
NPixels(int dim) const
{
  // Return number of pixels in dimension
  assert((dim >= 0) && (dim < 2));
  return (dim == 0) ? width : height;
}



inline RNScalar RGBDSurface::
PixelChannelValue(int ix, int iy, int channel_index) const
{
  // Return value in channel at pixel position
  return channels[channel_index]->GridValue(ix, iy);
}



inline RNScalar RGBDSurface::
PixelChannelValue(const R2Point& image_position, int channel_index) const
{
  // Return value in channel at pixel position
  return channels[channel_index]->GridValue(image_position);
}



inline R3Point RGBDSurface::
WorldOrigin(void) const
{
  // Return origin of planar region
  return surface_to_world.Matrix() * R3zero_point;
}



inline R3Vector RGBDSurface::
WorldNormal(void) const
{
  // Return normal vector of planar region
  return surface_to_world.Matrix() * R3posz_vector;
}



inline R3Vector RGBDSurface::
WorldAxis(int dimension) const
{
  // Return axis vector of planar region
  return surface_to_world.Matrix() * R3xyz_triad[dimension];
}



inline RNLength RGBDSurface::
WorldLength(int dimension) const
{
  // Return length of box in dimension
  if (dimension == RN_X) return resolution * width;
  else if (dimension == RN_Y) return resolution *height;
  else return 0;
}



inline R3Point RGBDSurface::
WorldCentroid(void) const
{
  // Return centroid of surface
  return WorldOrigin() + 0.5*WorldLength(RN_X)*WorldAxis(RN_X) + 0.5*WorldLength(RN_Y)*WorldAxis(RN_Y);
}



inline R3Box RGBDSurface::
WorldBBox(void) const
{
  // Return bounding box
  R3Box bbox = R3null_box;
  R3Point origin = WorldOrigin();
  bbox.Union(origin);
  bbox.Union(origin + WorldAxis(RN_X)*WorldLength(RN_X));
  bbox.Union(origin + WorldAxis(RN_Y)*WorldLength(RN_Y));
  bbox.Union(origin + WorldAxis(RN_X)*WorldLength(RN_X) + WorldAxis(RN_Y)*WorldLength(RN_Y));
  return bbox;;
}



inline RNArea RGBDSurface::
WorldArea(void) const
{
  // Return area of surface
  return WorldLength(RN_X) * WorldLength(RN_Y);
}



inline const R3Affine& RGBDSurface::
SurfaceToWorld(void) const
{
  // Return transformation that transforms surface coordinates into world coordinates
  return surface_to_world;
}


  
inline RNLength RGBDSurface::
Resolution(void) const
{
  // Return size of a pixel in surface coordinates
  // This is the scale factor that takes texture coordinates to surface coordinates 
  return resolution;
}



inline const char *RGBDSurface::
TextureFilename(void) const
{
  // Return color filename
  return texture_filename;
}



