////////////////////////////////////////////////////////////////////////
// Source file for RGBD coordinate transformation functions
////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////
// Include files
////////////////////////////////////////////////////////////////////////

#include "RGBD.h"



////////////////////////////////////////////////////////////////////////
// Transformation Functions
////////////////////////////////////////////////////////////////////////

int RGBDTransformTextureToSurface(const R2Point& texture_position, R2Point& surface_position, const RGBDSurface *surface)
{
  // Transform from position in texture coordinates to surface coordinates
  if (RNIsZero(surface->Resolution())) return 0;
  surface_position = surface->Resolution() * texture_position;
  return 1;
}



int RGBDTransformTextureToWorld(const R2Point& texture_position, R3Point& world_position, const RGBDSurface *surface)
{
  // Transform from position in texture coordinates to world coordinates
  R2Point surface_position;
  if (!RGBDTransformTextureToSurface(texture_position, surface_position, surface)) return 0;
  if (!RGBDTransformSurfaceToWorld(surface_position, world_position, surface)) return 0;
  return 1;
}



int RGBDTransformTextureToCamera(const R2Point& texture_position, R3Point& camera_position, const RGBDSurface *surface, const RGBDImage *image)
{
  // Transform from position in texture coordinates to camera coordinates
  R3Point world_position;
  if (!RGBDTransformTextureToWorld(texture_position, world_position, surface)) return 0;
  if (!RGBDTransformWorldToCamera(world_position, camera_position, image)) return 0;
  return 1;
}



int RGBDTransformTextureToImage(const R2Point& texture_position, R2Point& image_position, const RGBDSurface *surface, const RGBDImage *image)
{
  // Transform from position in texture coordinates to image coordinates
  R3Point world_position;
  if (!RGBDTransformTextureToWorld(texture_position, world_position, surface)) return 0;
  if (!RGBDTransformWorldToImage(world_position, image_position, image)) return 0;
  return 1;
}



int RGBDTransformSurfaceToTexture(const R2Point& surface_position, R2Point& texture_position, const RGBDSurface *surface)
{
  // Transform from position in surface coordinates to texture coordinates
  if (RNIsZero(surface->Resolution())) return 0;
  texture_position = surface_position / surface->Resolution();
  return 1;
}



int RGBDTransformSurfaceToWorld(const R2Point& surface_position, R3Point& world_position, const RGBDSurface *surface)
{
  // Transform from position in surface coordinates to world coordinates
  world_position.Reset(surface_position[0], surface_position[1], 0);
  world_position.Transform(surface->SurfaceToWorld());
  return 1;
}



int RGBDTransformSurfaceToCamera(const R2Point& surface_position, R3Point& camera_position, const RGBDSurface *surface, const RGBDImage *image)
{
  // Transform from position in surface coordinates to camera coordinates
  R3Point world_position;
  if (!RGBDTransformSurfaceToWorld(surface_position, world_position, surface)) return 0;
  if (!RGBDTransformWorldToCamera(world_position, camera_position, image)) return 0;
  return 1;
}



int RGBDTransformSurfaceToImage(const R2Point& surface_position, R2Point& image_position, const RGBDSurface *surface, const RGBDImage *image)
{
  // Transform from position in surface coordinates to image coordinates
  R3Point world_position;
  if (!RGBDTransformSurfaceToWorld(surface_position, world_position, surface)) return 0;
  if (!RGBDTransformWorldToImage(world_position, image_position, image)) return 0;
  return 1;
}



int RGBDTransformWorldToTexture(const R3Point& world_position, R2Point& texture_position, const RGBDSurface *surface)
{
  // Transform from position in world coordinates to texture coordinates
  R2Point surface_position;
  if (!RGBDTransformWorldToSurface(world_position, surface_position, surface)) return 0;
  if (!RGBDTransformSurfaceToTexture(surface_position, texture_position, surface)) return 0;
  return 1;

  // Return success
  return 1;
}



int RGBDTransformWorldToSurface(const R3Point& world_position, R2Point& surface_position, const RGBDSurface *surface)
{
  // Transform from position in world coordinates to surface coordinates
  R3Point tmp_position = world_position;
  tmp_position.InverseTransform(surface->SurfaceToWorld());
  surface_position.Reset(tmp_position[0], tmp_position[1]);
  return 1;
}



int RGBDTransformWorldToCamera(const R3Point& world_position, R3Point& camera_position, const RGBDImage *image)
{
  // Transform from position in world coordinates to camera coordinates
  camera_position = world_position;
  camera_position.InverseTransform(image->CameraToWorld());
  return 1;

  // Return success
  return 1;
}



int RGBDTransformWorldToImage(const R3Point& world_position, R2Point& image_position, const RGBDImage *image)
{
  // Transform from position in world coordinates to image coordinates
  R3Point camera_position;
  if (!RGBDTransformWorldToCamera(world_position, camera_position, image)) return 0;
  if (!RGBDTransformCameraToImage(camera_position, image_position, image)) return 0;
  return 1;

  // Return success
  return 1;
}



int RGBDTransformCameraToTexture(const R3Point& camera_position, R2Point& texture_position, const RGBDImage *image, const RGBDSurface *surface)
{
  // Transform from position in camera coordinates to texture coordinates
  R3Point world_position;
  if (!RGBDTransformCameraToWorld(camera_position, world_position, image)) return 0;
  if (!RGBDTransformWorldToTexture(world_position, texture_position, surface)) return 0;
  return 1;
}



int RGBDTransformCameraToSurface(const R3Point& camera_position, R2Point& surface_position, const RGBDImage *image, const RGBDSurface *surface)
{
  // Transform from position in camera coordinates to surface coordinates
  R3Point world_position;
  if (!RGBDTransformCameraToWorld(camera_position, world_position, image)) return 0;
  if (!RGBDTransformWorldToSurface(world_position, surface_position, surface)) return 0;
  return 1;
}



int RGBDTransformCameraToWorld(const R3Point& camera_position, R3Point& world_position, const RGBDImage *image)
{
  // Transform from camera coordinates to world coordinates
  world_position = camera_position;
  world_position.Transform(image->CameraToWorld());
  return 1;
}



int RGBDTransformCameraToImage(const R3Point& camera_position, R2Point& image_position, const RGBDImage *image)
{
  // Transform from position in camera coordinates to image coordinates

  // Get/check depth
  RNScalar depth = -camera_position[2];
  if ((RNIsZero(depth)) || (depth == R2_GRID_UNKNOWN_VALUE)) return 0;

  // Get/check intrinsics matrix
  const R3Matrix& intrinsics_matrix = image->Intrinsics();
  if (RNIsZero(intrinsics_matrix[0][0])) return 0;
  if (RNIsZero(intrinsics_matrix[1][1])) return 0;
  
  // Transform from position in image coordinates to camera coordinates (where camera is looking down -Z, up is +Y, right is +X)
  image_position[0] = intrinsics_matrix[0][2] + camera_position[0] * intrinsics_matrix[0][0] / depth;
  image_position[1] = intrinsics_matrix[1][2] + camera_position[1] * intrinsics_matrix[1][1] / depth;

  // Check image position
  if ((image_position[0] < 0) || (image_position[0] >= image->NPixels(RN_X))) return 0; 
  if ((image_position[1] < 0) || (image_position[1] >= image->NPixels(RN_Y))) return 0;

  // Check pixel position
  int image_ix = (int) (image_position.X() + 0.5);
  if ((image_ix < 0) || (image_ix >= image->NPixels(RN_X))) return 0;
  int image_iy = (int) (image_position.Y() + 0.5);
  if ((image_iy < 0) || (image_iy >= image->NPixels(RN_Y))) return 0;

  // Check depth
  RNScalar image_depth = image->PixelDepth(image_ix, image_iy);
  if ((image_depth == R2_GRID_UNKNOWN_VALUE) || (depth == 0)) return 0;

  // If depth is not at least 90% of image depth, then probably occluded
  if (image_depth < 0.9 * depth) return 0;
  
  // Return success
  return 1;
}




int RGBDTransformImageToTexture(const R2Point& image_position, R2Point& texture_position, const RGBDImage *image, const RGBDSurface *surface)
{
  // Transform from position in image coordinates to texture coordinates
  R3Point world_position;
  if (!RGBDTransformImageToWorld(image_position, world_position, image)) return 0;
  if (!RGBDTransformWorldToTexture(world_position, texture_position, surface)) return 0;
  return 1;
}



int RGBDTransformImageToSurface(const R2Point& image_position, R2Point& surface_position, const RGBDImage *image, const RGBDSurface *surface)
{
  // Transform from position in image coordinates to surface coordinates
  R3Point world_position;
  if (!RGBDTransformImageToWorld(image_position, world_position, image)) return 0;
  if (!RGBDTransformWorldToSurface(world_position, surface_position, surface)) return 0;
  return 1;
}



int RGBDTransformImageToWorld(const R2Point& image_position, R3Point& world_position, const RGBDImage *image)
{
  // Transform from position in image coordinates to world coordinates
  R3Point camera_position;
  if (!RGBDTransformImageToCamera(image_position, camera_position, image)) return 0;
  if (!RGBDTransformCameraToWorld(camera_position, world_position, image)) return 0;
  return 1;
}



int RGBDTransformImageToCamera(const R2Point& image_position, R3Point& camera_position, const RGBDImage *image)
{
  // Check image position
  int ix = (int) (image_position[0] + 0.5);
  if ((ix < 0) || (ix >= image->NPixels(RN_X))) return 0; 
  int iy = (int) (image_position[1] + 0.5);
  if ((iy < 0) || (iy >= image->NPixels(RN_Y))) return 0; 
  
  // Get/check depth (point sample to avoid interpolation)
  RNScalar depth = image->PixelDepth(ix, iy);
  if ((RNIsZero(depth)) || (depth == R2_GRID_UNKNOWN_VALUE)) return 0;

  // Get/check intrinsics matrix
  const R3Matrix& intrinsics_matrix = image->Intrinsics();
  if (RNIsZero(intrinsics_matrix[0][0])) return 0;
  if (RNIsZero(intrinsics_matrix[1][1])) return 0;
  
  // Transform from position in image coordinates to camera coordinates (where camera is looking down -Z, up is +Y, right is +X)
  camera_position[0] = (image_position[0] - intrinsics_matrix[0][2]) * depth / intrinsics_matrix[0][0];
  camera_position[1] = (image_position[1] - intrinsics_matrix[1][2]) * depth / intrinsics_matrix[1][1];
  camera_position[2] = -depth;

  // Return success
  return 1;
}




