// Source file for the rgbd loader program



////////////////////////////////////////////////////////////////////////
// Include files 
////////////////////////////////////////////////////////////////////////

#include "R3Graphics/R3Graphics.h"
#include "RGBD/RGBD.h"
#include "fglut/fglut.h"
#include "render.h"



////////////////////////////////////////////////////////////////////////
// Program arguments
////////////////////////////////////////////////////////////////////////

static const char *input_configuration_filename = NULL;
static const char *output_configuration_filename = NULL;
static int read_every_kth_image = 1;
static int write_every_kth_image = 1;
static double resolution = 0;
static int interactive = 0;
static int print_verbose = 0;



////////////////////////////////////////////////////////////////////////
// Internal variables
////////////////////////////////////////////////////////////////////////

// GLUT variables 

static int GLUTwindow = 0;
static int GLUTwindow_width = 640;
static int GLUTwindow_height = 480;
static int GLUTmouse[2] = { 0, 0 };
static int GLUTbutton[3] = { 0, 0, 0 };
static int GLUTmouse_drag = 0;
static int GLUTmodifiers = 0;


// Application variables

static RGBDConfiguration configuration;
static char *screenshot_image_name = NULL;
static int selected_surface_index = -1;
static int selected_image_index = -1;
static R2Point rubber_box_corners[2] = { R2Point(0,0), R2Point(0,0) };
static RNBoolean rubber_box_active = FALSE;
static R3Point center(0, 0, 0);
static R3Viewer viewer;



// Display variables

static int show_cameras = 1;
static int show_bboxes = 0;
static int show_images = 0;
static int show_points = 1;
static int show_planes = 0;
static int show_textures = 1;
static int show_axes = 0;



////////////////////////////////////////////////////////////////////////
// Read/Write functions
////////////////////////////////////////////////////////////////////////

static int
ReadConfiguration(RGBDConfiguration& configuration, const char *filename) 
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Reading configuration from %s ...\n", filename);
    fflush(stdout);
  }

  // Read file
  if (!configuration.ReadFile(filename, read_every_kth_image)) return 0;

  // Set surface resolutions if specified on command line
  if (resolution > 0) {
    for (int i = 0; i < configuration.NSurfaces(); i++) {
      RGBDSurface *surface = configuration.Surface(i);
      surface->SetResolution(resolution);
    }
  }

  // Print statistics
  if (print_verbose) {
    R3Box bbox = configuration.WorldBBox();
    R3Point centroid = configuration.WorldCentroid();
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Images = %d\n", configuration.NImages());
    printf("  # Surfaces = %d\n", configuration.NSurfaces());
    printf("  Centroid = %g %g %g\n", centroid.X(), centroid.Y(), centroid.Z());
    printf("  BBox = %g %g %g  %g %g %g\n", bbox[0][0], bbox[0][1], bbox[0][2], bbox[1][0], bbox[1][1], bbox[1][2]);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WriteConfiguration(RGBDConfiguration& configuration, const char *filename) 
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Writing configuration to %s ...\n", filename);
    fflush(stdout);
  }

  // Write file
  if (!configuration.WriteFile(filename, write_every_kth_image)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Images = %d\n", configuration.NImages());
    printf("  # Surfaces = %d\n", configuration.NSurfaces());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Draw functions
////////////////////////////////////////////////////////////////////////

static void
DrawCameras(int color_scheme = RGBD_INDEX_COLOR_SCHEME)
{
  // Draw cameras
  for (int i = 0; i < configuration.NImages(); i++) {
    RGBDImage *image = configuration.Image(i);
    int c = (i == selected_image_index) ? RGBD_HIGHLIGHT_COLOR_SCHEME : color_scheme;
    image->DrawCamera(c);
  }
}



static void
DrawImages(int color_scheme = RGBD_PHOTO_COLOR_SCHEME)
{
  // Draw images
  for (int i = 0; i < configuration.NImages(); i++) {
    RGBDImage *image = configuration.Image(i);
    int c = (i == selected_image_index) ? RGBD_HIGHLIGHT_COLOR_SCHEME : color_scheme;
    if (i == selected_image_index) image->DrawImage(c);
  }
}



static void
DrawPoints(int color_scheme = RGBD_PHOTO_COLOR_SCHEME, int skip = 10)
{
  // Draw pixels of all images as points in world space
  for (int i = 0; i < configuration.NImages(); i++) {
    RGBDImage *image = configuration.Image(i);
    int c = (i == selected_image_index) ? RGBD_HIGHLIGHT_COLOR_SCHEME : color_scheme;
    image->DrawPoints(c, skip);
  }
}



static void
DrawBBoxes(int color_scheme = RGBD_INDEX_COLOR_SCHEME)
{
  // Draw pixels of all images as bboxs in world space
  for (int i = 0; i < configuration.NImages(); i++) {
    RGBDImage *image = configuration.Image(i);
    int c = (i == selected_image_index) ? RGBD_HIGHLIGHT_COLOR_SCHEME : color_scheme;
    image->DrawBBox(c);
  }
}



static void
DrawPlanes(int color_scheme = RGBD_INDEX_COLOR_SCHEME)
{
  // Draw surfaces
  for (int i = 0; i < configuration.NSurfaces(); i++) {
    RGBDSurface *surface = configuration.Surface(i);
    int c = (i == selected_surface_index) ? RGBD_HIGHLIGHT_COLOR_SCHEME : color_scheme;
    surface->DrawPlane(c);
  }
}



static void
DrawTextures(int color_scheme = RGBD_PHOTO_COLOR_SCHEME)
{
  // Draw surfaces
  for (int i = 0; i < configuration.NSurfaces(); i++) {
    RGBDSurface *surface = configuration.Surface(i);
    int c = (i == selected_surface_index) ? RGBD_HIGHLIGHT_COLOR_SCHEME : color_scheme;
    surface->DrawTexture(c);
  }
}



static void
DrawAxes(void)
{
  // Draw axes 
  RNScalar d = 1;
  glDisable(GL_LIGHTING);
  glLineWidth(3);
  R3BeginLine();
  glColor3f(1, 0, 0);
  R3LoadPoint(R3zero_point + 0.5 * d * R3negx_vector);
  R3LoadPoint(R3zero_point + d * R3posx_vector);
  R3EndLine();
  R3BeginLine();
  glColor3f(0, 1, 0);
  R3LoadPoint(R3zero_point + 0.5 * d * R3negy_vector);
  R3LoadPoint(R3zero_point + d * R3posy_vector);
  R3EndLine();
  R3BeginLine();
  glColor3f(0, 0, 1);
  R3LoadPoint(R3zero_point + 0.5 * d * R3negz_vector);
  R3LoadPoint(R3zero_point + d * R3posz_vector);
  R3EndLine();
  glLineWidth(1);
}



static void 
DrawRubberBox(void)
{
  // Set rendering modes
  glDrawBuffer(GL_FRONT);
  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);
  glDepthMask(0);
  glLineStipple(1, 0xFF00);
  glEnable(GL_LINE_STIPPLE);
  glLogicOp(GL_XOR);
  glEnable(GL_COLOR_LOGIC_OP);

  // Set projection matrix
  glMatrixMode(GL_PROJECTION);  
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(0, GLUTwindow_width, 0, GLUTwindow_height);

  // Set model view matrix
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  // Draw box
  glBegin(GL_LINE_LOOP);
  glColor3f(1.0, 1.0, 1.0);
  glVertex2f(rubber_box_corners[0][0], rubber_box_corners[0][1]);
  glVertex2f(rubber_box_corners[0][0], rubber_box_corners[1][1]);
  glVertex2f(rubber_box_corners[1][0], rubber_box_corners[1][1]);
  glVertex2f(rubber_box_corners[1][0], rubber_box_corners[0][1]);
  glEnd();

  // Reset projection matrix
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  // Reset model view matrix
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  // Reset rendering modes
  glDisable(GL_COLOR_LOGIC_OP);
  glDisable(GL_LINE_STIPPLE);
  glEnable(GL_DEPTH_TEST);
  glDepthMask(1);
  glDrawBuffer(GL_BACK);
  glFlush();
}



////////////////////////////////////////////////////////////////////////
// Selection with cursor
////////////////////////////////////////////////////////////////////////

static int
Pick(int x, int y,
  RGBDSurface **picked_surface = NULL, RGBDImage **picked_image = NULL,
  R3Point *picked_position = NULL, int pick_tolerance = 10)
{
  // Initialize pick results
  int pick_result = 0;
  if (picked_surface) *picked_surface = NULL;
  if (picked_image) *picked_image = NULL;
  if (picked_position) *picked_position = R3zero_point;

  // Draw everything
  viewer.Load();
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glDisable(GL_LIGHTING);
  glPointSize(pick_tolerance);
  glLineWidth(pick_tolerance);
  if (picked_image || !picked_surface) {
    if (show_cameras) DrawCameras(RGBD_INDEX_COLOR_SCHEME);
    if (show_points) DrawPoints(RGBD_INDEX_COLOR_SCHEME);
  }
  if (picked_surface || !picked_image) {
    if (show_planes || show_textures) DrawPlanes(RGBD_INDEX_COLOR_SCHEME);
  }
  glPointSize(1.0);
  glLineWidth(1.0);
  glFinish();

  // Read color buffer at cursor position
  unsigned char rgba[4];
  glReadPixels(x, y, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, rgba);
  int r = rgba[0] & 0xFF;
  int g = rgba[1] & 0xFF;
  int b = rgba[2] & 0xFF;
  // int a = rgba[3] & 0xFF;

  // Figure out what was picked
  if (r > 0) {
    // Picked surface
    int surface_index = (int) ((configuration.NSurfaces()*r/255.0) + 0.5) - 1;
    if ((surface_index < 0) || (surface_index >= configuration.NSurfaces())) return 0;
    if (picked_surface) *picked_surface = configuration.Surface(surface_index);
    pick_result = RGBD_SURFACE_SELECTION;
  }
  else if ((g > 0) || (b > 0)) {
    // Picked image
    int image_index = (int) ((configuration.NImages()*((g << 8) | b) / 65535.0) + 0.5) - 1;
    if ((image_index < 0) || (image_index >= configuration.NImages())) return 0;
    if (picked_image) *picked_image = configuration.Image(image_index);
    pick_result = RGBD_IMAGE_SELECTION;
  }
  else {
    // Picked background
    return RGBD_NO_SELECTION;
  }

  // Return position
  if (picked_position) {
    // Find hit position
    GLfloat depth;
    GLdouble p[3];
    GLint viewport[4];
    GLdouble modelview_matrix[16];
    GLdouble projection_matrix[16];
    glGetIntegerv(GL_VIEWPORT, viewport);
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
    glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);
    gluUnProject(x, y, depth, modelview_matrix, projection_matrix, viewport, &(p[0]), &(p[1]), &(p[2]));
    R3Point position(p[0], p[1], p[2]);
    *picked_position = position;
  }

  // Return success
  return pick_result;
}



static int
ProcessRubberBox(void)
{
  // Create surface ...
  
  // Create rubber box (with sorted corners)
  R2Box rubber_box = R2null_box;
  rubber_box.Union(rubber_box_corners[0]);
  rubber_box.Union(rubber_box_corners[1]);
  if ((rubber_box.XLength() < 10) || (rubber_box.YLength() < 10)) {
    printf("Box is too small -- could not create surface\n");
    fflush(stdout);
    return 0;
  }
   
  // Create random set of points projecting inside rubber box
  int npoints = 0;
  int max_npoints = 1024;
  R3Point *points = new R3Point [ max_npoints ];
  for (int i = 0; i < 10 * max_npoints; i++) {
    int image_index = (int) (RNRandomScalar() * configuration.NImages());
    RGBDImage *image = configuration.Image(image_index);
    double x = RNRandomScalar() * image->NPixels(RN_X);
    double y = RNRandomScalar() * image->NPixels(RN_Y);
    R3Point world_position;
    if (!RGBDTransformImageToWorld(R2Point(x, y), world_position, image)) continue;
    R2Point viewport_position = viewer.ViewportPoint(world_position);
    if (!R2Contains(rubber_box, viewport_position)) continue;
    points[npoints++] = world_position;
    if (npoints >= max_npoints) break;
  }
  
  // Check number of points
  if (npoints < 3) {
    printf("Did not find enough points inside box -- could not create surface\n");
    delete [] points;
    fflush(stdout);
    return 0;
  }
  
  // Compute centroid and principle axes of points
  R3Point centroid = R3Centroid(npoints, points);
  R3Triad axes = R3PrincipleAxes(centroid, npoints, points);
  if (axes[2].Dot(viewer.Camera().Towards()) > 0) {
    axes = R3Triad(axes[0], -axes[1], -axes[2]);
  }
  
  // Compute axis vectors
  R3Vector normal = axes[2]; normal.Normalize();
  R3Vector up = viewer.Camera().Up();
  R3Vector right = normal % up; right.Normalize();
  up = right % normal; up.Normalize();
  
  // Compute radii of points
  RNScalar radius[2] = { 0, 0 };
  for (int i = 0; i < npoints; i++) {
    RNScalar dx = fabs((points[i] - centroid).Dot(right));
    RNScalar dy = fabs((points[i] - centroid).Dot(up));
    if (dx > radius[0]) radius[0] = dx;
    if (dy > radius[1]) radius[1] = dy;
  }

  // Delete points
  delete [] points;

  // Compute resolution
  RNScalar res = (resolution > 0) ? resolution : configuration.AverageSurfaceResolution();

  // Create surface
  RGBDSurface *surface = new RGBDSurface(NULL, centroid, normal, up, radius, res);
  if (!surface) {
    printf("Ran out of memory -- could not create surface\n");
    fflush(stdout);
    return 0;
  }

  // Insert surface into configuration
  configuration.InsertSurface(surface);

  // Compute surface texture
  RGBDComputeTexture(surface);

  // Print message
  printf("Created surface %d\n", configuration.NSurfaces()-1);
  fflush(stdout);

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Utility functions
////////////////////////////////////////////////////////////////////////

static void
ResetViewer(void)
{
  // Get bounding box
  R3Box bbox = configuration.WorldBBox();

  // Restrict bounding box only to surfaces, if there are any
  if (configuration.NSurfaces() > 0) {
    bbox = R3null_box;
    for (int i = 0; i < configuration.NSurfaces(); i++) {
      RGBDSurface *surface = configuration.Surface(i);
      bbox.Union(surface->WorldBBox());
    }
  }

  // Initialize viewing center
  center = bbox.Centroid();

  // Initialize viewer
  RNLength r = bbox.DiagonalRadius();
  R3Point eye = center - R3posy_vector * (2.5 * r);
  R3Camera camera(eye, R3posy_vector, R3posz_vector, 0.4, 0.4, 0.01 * r, 100.0 * r);
  R2Viewport viewport(0, 0, GLUTwindow_width, GLUTwindow_height);
  viewer.SetViewport(viewport);
  viewer.SetCamera(camera);
}



static void 
AlignWithAxes(void)
{
  // Check images and surfaces
  if (configuration.NImages() == 0) return;
  
  // Compute posz = average image up vector
  R3Vector posz = R3zero_vector;
  for (int i = 0; i < configuration.NImages(); i++) {
    RGBDImage *image = configuration.Image(i);
    posz += image->WorldUp();
  }

  // Compute origin and posy
  R3Point origin = R3zero_point;
  R3Vector posy = R3zero_vector;
  if (configuration.NSurfaces() > 0) {
    // Find surface with largest area
    RNArea largest_area = 0;
    for (int i = 0; i < configuration.NSurfaces(); i++) {
      RGBDSurface *surface = configuration.Surface(i);
      RNArea area = surface->WorldArea();
      if (area > largest_area) {
        origin = surface->WorldCentroid();
        posy = -(surface->WorldNormal());
        largest_area = area;
      }
    }
  }
  else {
    // Average positions and normals of pixels
    int count = 0;
    origin = R3zero_point;
    for (int i = 0; i < configuration.NImages(); i++) {
      RGBDImage *image = configuration.Image(i);
      int ix = image->NPixels(RN_X)/2;
      int iy = image->NPixels(RN_Y)/2;
      R3Point world_position;
      if (RGBDTransformImageToWorld(R2Point(ix, iy), world_position, image)) {
        posy += -(image->PixelWorldNormal(ix, iy));
        origin += world_position;
        count++;
      }
    }

    // Compute averages
    if (count > 0) origin /= count;
  }

  // Compute orthogonal triad of axes
  posy.Normalize();
  posz.Normalize();
  R3Vector posx = posy % posz; posx.Normalize();
  posz = posx % posy; posz.Normalize();
  R3Triad axes(posx, posy, posz);
  
  // Compute transformation
  R3CoordSystem cs(origin, axes);
  R3Affine transformation(cs.InverseMatrix(), 0);

  // Apply transformation
  configuration.Transform(transformation);
}



static int
DeleteSelectedSurface(void)
{
  // Print message
  if (selected_surface_index >= 0) {
    printf("Deleted surface %d\n", selected_surface_index);
    fflush(stdout);
  }
  else {
    printf("Select a surface first, and then type ctrl-d if you want to delete it\n");
    fflush(stdout);
    return 0;
  }

  // Delete selected surface
  RGBDSurface *selected_surface = configuration.Surface(selected_surface_index);
  delete selected_surface;

  // Reset selected surface index
  selected_surface_index = -1;

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// MAIN PROCESSING FUNCTION
////////////////////////////////////////////////////////////////////////

static int
ComputeTextures(RGBDConfiguration& configuration)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  if (print_verbose) {
    printf("Computing textures ...\n");
    fflush(stdout);
  }

  // Update textures
  if (!RGBDComputeTextures(&configuration)) return 0;

  // Print statistics
  if (print_verbose) {
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Surfaces = %d\n", configuration.NSurfaces());
    fflush(stdout);
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// GLUT user interface functions
////////////////////////////////////////////////////////////////////////

void GLUTStop(void)
{
  // Write configuration
  if (output_configuration_filename) {
    if (!WriteConfiguration(configuration, output_configuration_filename)) exit(-1);
  }

  // Destroy window 
  glutDestroyWindow(GLUTwindow);

  // Exit
  exit(0);
}



void GLUTRedraw(void)
{
  // Set viewing transformation
  viewer.Camera().Load();

  // Clear window 
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Draw everything
  if (show_cameras) DrawCameras();
  if (show_bboxes) DrawBBoxes();
  if (show_images) DrawImages();
  if (show_planes) DrawPlanes();
  if (show_points) DrawPoints();
  if (show_textures) DrawTextures();
  if (show_axes) DrawAxes();

  // Capture screenshot image 
  if (screenshot_image_name) {
    if (print_verbose) printf("Creating image %s\n", screenshot_image_name);
    R2Image image(GLUTwindow_width, GLUTwindow_height, 3);
    image.Capture();
    image.Write(screenshot_image_name);
    screenshot_image_name = NULL;
  }

  // Swap buffers 
  glutSwapBuffers();
}    



void GLUTResize(int w, int h)
{
  // Resize window
  glViewport(0, 0, w, h);

  // Resize viewer viewport
  viewer.ResizeViewport(0, 0, w, h);

  // Remember window size 
  GLUTwindow_width = w;
  GLUTwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUTMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUTmouse[0];
  int dy = y - GLUTmouse[1];
  
  // Update mouse drag
  GLUTmouse_drag += dx*dx + dy*dy;

  // Check if control key is down
  if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
    // Update rubber box
    if (rubber_box_active) {
      DrawRubberBox();
      rubber_box_corners[1] = R2Point(x, y);
      DrawRubberBox();
    }
    else {
      rubber_box_active = TRUE;
      rubber_box_corners[0] = R2Point(x, y);
      rubber_box_corners[1] = R2Point(x, y);
      DrawRubberBox();
    }
  }
  else {
    // World in hand navigation 
    if (GLUTbutton[0]) viewer.RotateWorld(1.0, center, x, y, dx, dy);
    else if (GLUTbutton[1]) viewer.ScaleWorld(1.0, center, x, y, dx, dy);
    else if (GLUTbutton[2]) viewer.TranslateWorld(1.0, center, x, y, dx, dy);
    if (GLUTbutton[0] || GLUTbutton[1] || GLUTbutton[2]) glutPostRedisplay();
  }

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;
}



void GLUTMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Mouse is going down
  if (state == GLUT_DOWN) {
    // Reset mouse drag
    GLUTmouse_drag = 0;

    // Process thumbwheel
    if (button == 3) viewer.ScaleWorld(center, 0.9);
    else if (button == 4) viewer.ScaleWorld(center, 1.1);
  }
  else if (button == 0) {
    // Check for double click  
    static RNBoolean double_click = FALSE;
    static RNTime last_mouse_up_time;
    double_click = (!double_click) && (last_mouse_up_time.Elapsed() < 0.4);
    last_mouse_up_time.Read();

    // Check for click (rather than drag)
    if (GLUTmouse_drag < 100) {
      // Reset selected indices
      selected_surface_index = -1;
      selected_image_index = -1;

      // Select image or surface
      R3Point selected_position(0,0,0);
      RGBDImage *selected_image = NULL;
      RGBDSurface *selected_surface = NULL;
      if (Pick(x, y, &selected_surface, &selected_image, &selected_position)) {
        // Remember selected indices
        if (selected_surface) selected_surface_index = selected_surface->ConfigurationIndex();
        if (selected_image) selected_image_index = selected_image->ConfigurationIndex();

        // Print message
        if (selected_surface) {
          printf("Selected surface %d at %g %g %g\n", selected_surface->ConfigurationIndex(),
            selected_position.X(), selected_position.Y(), selected_position.Z());
        }
        if (selected_image) {
          printf("Selected image %d and %g %g %g\n", selected_image->ConfigurationIndex(),
            selected_position.X(), selected_position.Y(), selected_position.Z());
        }
        
        // Set viewing center of rotation and scale if double-click
        if (double_click) center = selected_position;
      }
    }
    else {
      // Process rubber box 
      if (rubber_box_active) {
        DrawRubberBox();
        ProcessRubberBox();
        rubber_box_corners[0] = R2zero_point;
        rubber_box_corners[1] = R2zero_point;
        rubber_box_active = FALSE;
      }
    }
  }

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUTbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

   // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUTSpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Process keyboard button event 
  switch (key) {
  case GLUT_KEY_F1:
    AlignWithAxes();
    ResetViewer();
    break;

  case GLUT_KEY_LEFT:
  case GLUT_KEY_RIGHT:
    // Select next image
    if (configuration.NImages() > 0) {
      if (key == GLUT_KEY_RIGHT) selected_image_index++;
      else if (key == GLUT_KEY_LEFT) selected_image_index--;
      if (selected_image_index < 0) selected_image_index = 0;
      if (selected_image_index >= configuration.NImages()) selected_image_index = configuration.NImages()-1;
      RGBDImage *selected_image = configuration.Image(selected_image_index);
      R3Point viewpoint = selected_image->WorldViewpoint();
      R3Vector towards = selected_image->WorldTowards();
      R3Vector up = selected_image->WorldUp();
      RNAngle xfov = selected_image->XFov();
      RNAngle yfov = selected_image->YFov();
      R3Camera camera(viewpoint, towards, up, xfov, yfov, 0.01, 100.0);
      viewer.SetCamera(camera);
      center = viewpoint + towards;
    }
    break;
  }

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Redraw
  glutPostRedisplay();
}



void GLUTKeyboard(unsigned char key, int x, int y)
{
  // Process keyboard button event 
  switch (key) {
  case '~': {
    // Dump screen shot to file iX.jpg
    static char buffer[64];
    static int image_count = 1;
    sprintf(buffer, "i%d.jpg", image_count++);
    screenshot_image_name = buffer;
    break; }

  case 4: // ctrl-D
    DeleteSelectedSurface();
    break;

  case 18: // ctrl-R
    ResetViewer();
    break;

  case 'A':
  case 'a':
    show_axes = !show_axes;
    break;

  case 'B':
  case 'b':
    show_bboxes = !show_bboxes;
    break;

  case 'C':
  case 'c':
    show_cameras = !show_cameras;
    break;

  case 'I':
  case 'i':
    show_images = !show_images;
    break;

  case 'P':
  case 'p':
    show_points = !show_points;
    break;

  case 'S':
  case 's':
    show_planes = !show_planes;
    break;

  case 'T':
  case 't':
    show_textures = !show_textures;
    break;

  case ' ':
    ComputeTextures(configuration);
    break;

  case 27: // ESCAPE
    GLUTStop();
    break;
  }

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = GLUTwindow_height - y;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Redraw
  glutPostRedisplay();  
}




void GLUTInterface(void)
{
  // Open window
  int argc = 0;
  char *argv[1];
  argv[argc++] = strdup("conf2texture");
  glutInit(&argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUTwindow_width, GLUTwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUTwindow = glutCreateWindow("Configuration Viewer");

  // Initialize graphics modes  
  glEnable(GL_DEPTH_TEST);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUTRedraw);
  glutReshapeFunc(GLUTResize);
  glutKeyboardFunc(GLUTKeyboard);
  glutSpecialFunc(GLUTSpecial);
  glutMouseFunc(GLUTMouse);
  glutMotionFunc(GLUTMotion);

  // Reset viewer
  ResetViewer();

  // Run main loop -- never returns 
  glutMainLoop();
}


 
////////////////////////////////////////////////////////////////////////
// PROGRAM ARGUMENT PARSING
////////////////////////////////////////////////////////////////////////

static int 
ParseArgs(int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1;
      else if (!strcmp(*argv, "-interactive")) interactive = 1;
      else if (!strcmp(*argv, "-read_every_kth_image")) { argc--; argv++; read_every_kth_image = atoi(*argv); }
      else if (!strcmp(*argv, "-write_every_kth_image")) { argc--; argv++; write_every_kth_image = atoi(*argv); }
      else if (!strcmp(*argv, "-resolution")) { argc--; argv++; resolution = atof(*argv); }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
    else {
      if (!input_configuration_filename) input_configuration_filename = *argv;
      else if (!output_configuration_filename) output_configuration_filename = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
  }

  // Check filenames
  if (!input_configuration_filename) {
    fprintf(stderr, "Usage: texture inputconfigurationfile [outputconfigurationfile] [options]\n");
    return 0;
  }

  // If there is no output file, then start interactive interface
  if (!output_configuration_filename) interactive = TRUE;

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// MAIN
////////////////////////////////////////////////////////////////////////

int
main(int argc, char **argv)
{
  // Check number of arguments
  if (!ParseArgs(argc, argv)) exit(1);

  // Read configuration
  if (!ReadConfiguration(configuration, input_configuration_filename)) exit(-1);

  // Compute textures
  if (!ComputeTextures(configuration)) exit(-1);

  // Check if interactive
  if (interactive) {
    // Begin viewing interface
    GLUTInterface();
  }
  else {
    // Write configuration
    if (output_configuration_filename) {
      if (!WriteConfiguration(configuration, output_configuration_filename)) exit(-1);
    }
  }
  
  // Return success 
  return 0;
}



